diff --git a/gdb/ChangeLog b/gdb/ChangeLog
index 2810fff012..0e79bd0e6b 100644
--- a/gdb/ChangeLog
+++ b/gdb/ChangeLog
@@ -1,4 +1,21 @@
-2019-05-14  Alan Hayward  <alan.hayward@arm.com>
+2019-05-17  Alan Hayward  <alan.hayward@arm.com>
+
+       * cli/cli-interp.c (struct saved_output_files): Add saved entry.
+       (cli_interp_base::set_logging): Check debug_redirect.
+       * cli/cli-interp.h (set_logging): Add debug_redirect parameter.
+       * cli/cli-logging.c (debug_redirect): Add static variable.
+       (pop_output_files): Add default param.
+       (handle_redirections): Print debug setting.
+       (show_logging_command): Likewise.
+       (_initialize_cli_logging): Add debugredirect command.
+       * interps.c (current_interp_set_logging): Add debug_redirect
+       parameter.
+       * interps.h (set_logging): Add debug_redirect parameter.
+       (current_interp_set_logging): Likewise.
+       * mi/mi-common.h: Likewise.
+       * mi/mi-interp.c (mi_interp::set_logging): Likewise.
+
+2019-05-17  Alan Hayward  <alan.hayward@arm.com>
            Tom Tromey  <tromey@adacore.com>

        * cli/cli-interp.c (cli_interp_base::set_logging): Create tee_file
diff --git a/gdb/cli/cli-interp.c b/gdb/cli/cli-interp.c
index 7876f91080..a8cf6172c1 100644
--- a/gdb/cli/cli-interp.c
+++ b/gdb/cli/cli-interp.c
@@ -395,13 +395,15 @@ struct saved_output_files
   ui_file *log;
   ui_file *targ;
   ui_file *targerr;
+  ui_file *file_to_delete;
 };
 static saved_output_files saved_output;

 /* See cli-interp.h.  */

 void
-cli_interp_base::set_logging (ui_file_up logfile, bool logging_redirect)
+cli_interp_base::set_logging (ui_file_up logfile, bool logging_redirect,
+                             bool debug_redirect)
 {
   if (logfile != nullptr)
     {
@@ -411,28 +413,34 @@ cli_interp_base::set_logging (ui_file_up logfile, bool logging_redirect)
       saved_output.targ = gdb_stdtarg;
       saved_output.targerr = gdb_stdtargerr;

-      /* If logging is being redirected, then grab logfile.  */
+      /* If something is being redirected, then grab logfile.  */
       ui_file *logfile_p = nullptr;
-      if (logging_redirect)
-       logfile_p = logfile.release ();
+      if (logging_redirect || debug_redirect)
+       {
+         logfile_p = logfile.get ();
+         saved_output.file_to_delete = logfile_p;
+       }

-      /* If logging is not being redirected, then a tee containing both the
+      /* If something is not being redirected, then a tee containing both the
         logfile and stdout.  */
       ui_file *tee = nullptr;
-      if (!logging_redirect)
-       tee = new tee_file (gdb_stdout, std::move (logfile));
+      if (!logging_redirect || !debug_redirect)
+       {
+         tee = new tee_file (gdb_stdout, std::move (logfile));
+         saved_output.file_to_delete = tee;
+       }

       gdb_stdout = logging_redirect ? logfile_p : tee;
-      gdb_stdlog = logging_redirect ? logfile_p : tee;
+      gdb_stdlog = debug_redirect ? logfile_p : tee;
       gdb_stderr = logging_redirect ? logfile_p : tee;
       gdb_stdtarg = logging_redirect ? logfile_p : tee;
       gdb_stdtargerr = logging_redirect ? logfile_p : tee;
     }
   else
     {
-      /* Only delete one of the files -- they are all set to the same
-        value.  */
-      delete gdb_stdout;
+      /* Delete the correct file.  If it's the tee then the logfile will also
+        be deleted.  */
+      delete saved_output.file_to_delete;

       gdb_stdout = saved_output.out;
       gdb_stderr = saved_output.err;
@@ -445,6 +453,7 @@ cli_interp_base::set_logging (ui_file_up logfile, bool logging_redirect)
       saved_output.log = nullptr;
       saved_output.targ = nullptr;
       saved_output.targerr = nullptr;
+      saved_output.file_to_delete = nullptr;
     }
 }

diff --git a/gdb/cli/cli-interp.h b/gdb/cli/cli-interp.h
index 0c2e73b6b3..334a37accf 100644
--- a/gdb/cli/cli-interp.h
+++ b/gdb/cli/cli-interp.h
@@ -28,7 +28,8 @@ public:
   explicit cli_interp_base (const char *name);
   virtual ~cli_interp_base () = 0;

-  void set_logging (ui_file_up logfile, bool logging_redirect) override;
+  void set_logging (ui_file_up logfile, bool logging_redirect,
+                   bool debug_redirect) override;
   void pre_command_loop () override;
   bool supports_command_editing () override;
 };
diff --git a/gdb/cli/cli-logging.c b/gdb/cli/cli-logging.c
index 670e7e2490..bef5f3939b 100644
--- a/gdb/cli/cli-logging.c
+++ b/gdb/cli/cli-logging.c
@@ -62,6 +62,7 @@ show_logging_overwrite (struct ui_file *file, int from_tty,

 /* Value as configured by the user.  */
 static int logging_redirect;
+static int debug_redirect;

 static void
 set_logging_redirect (const char *args,
@@ -81,7 +82,7 @@ show_logging_redirect (struct ui_file *file, int from_tty,
 static void
 pop_output_files (void)
 {
-  current_interp_set_logging (NULL, false);
+  current_interp_set_logging (NULL, false, false);

   /* Stay consistent with handle_redirections.  */
   if (!current_uiout->is_mi_like_p ())
@@ -112,12 +113,20 @@ handle_redirections (int from_tty)
       else
        fprintf_unfiltered (gdb_stdout, "Redirecting output to %s.\n",
                            logging_filename);
+
+      if (!debug_redirect)
+       fprintf_unfiltered (gdb_stdout, "Copying debug output to %s.\n",
+                           logging_filename);
+      else
+       fprintf_unfiltered (gdb_stdout, "Redirecting debug output to %s.\n",
+                           logging_filename);
     }

   saved_filename = xstrdup (logging_filename);

   /* Let the interpreter do anything it needs.  */
-  current_interp_set_logging (std::move (log), logging_redirect);
+  current_interp_set_logging (std::move (log), logging_redirect,
+                             debug_redirect);

   /* Redirect the current ui-out object's output to the log.  Use
      gdb_stdout, not log, since the interpreter may have created a tee
@@ -185,6 +194,11 @@ show_logging_command (const char *args, int from_tty)
     printf_unfiltered (_("Output will be sent only to the log file.\n"));
   else
     printf_unfiltered (_("Output will be logged and displayed.\n"));
+
+  if (debug_redirect)
+    printf_unfiltered (_("Debug output will be sent only to the log file.\n"));
+  else
+    printf_unfiltered (_("Debug output will be logged and displayed.\n"));
 }

 void
@@ -213,6 +227,15 @@ If redirect is on, output will go only to the log file."),
                           set_logging_redirect,
                           show_logging_redirect,
                           &set_logging_cmdlist, &show_logging_cmdlist);
+  add_setshow_boolean_cmd ("debugredirect", class_support,
+                          &debug_redirect, _("\
+Set the logging debug output mode."), _("\
+Show the logging debug output mode."), _("\
+If debug redirect is off, debug will go to both the screen and the log file.\n\
+If debug redirect is on, debug will go only to the log file."),
+                          set_logging_redirect,
+                          show_logging_redirect,
+                          &set_logging_cmdlist, &show_logging_cmdlist);
   add_setshow_filename_cmd ("file", class_support, &logging_filename, _("\
 Set the current logfile."), _("\
 Show the current logfile."), _("\
diff --git a/gdb/interps.c b/gdb/interps.c
index dfd6eb9065..372bd73bf8 100644
--- a/gdb/interps.c
+++ b/gdb/interps.c
@@ -255,13 +255,13 @@ set_top_level_interpreter (const char *name)
 }

 void
-current_interp_set_logging (ui_file_up logfile,
-                           bool logging_redirect)
+current_interp_set_logging (ui_file_up logfile, bool logging_redirect,
+                           bool debug_redirect)
 {
   struct ui_interp_info *ui_interp = get_current_interp_info ();
   struct interp *interp = ui_interp->current_interpreter;

-  interp->set_logging (std::move (logfile), logging_redirect);
+  interp->set_logging (std::move (logfile), logging_redirect, debug_redirect);
 }

 /* Temporarily overrides the current interpreter.  */
diff --git a/gdb/interps.h b/gdb/interps.h
index 1bdc56c839..7cdb4d97e8 100644
--- a/gdb/interps.h
+++ b/gdb/interps.h
@@ -61,7 +61,8 @@ public:
   /* Provides a hook for interpreters to do any additional
      setup/cleanup that they might need when logging is enabled or
      disabled.  */
-  virtual void set_logging (ui_file_up logfile, bool logging_redirect) = 0;
+  virtual void set_logging (ui_file_up logfile, bool logging_redirect,
+                           bool debug_redirect) = 0;

   /* Called before starting an event loop, to give the interpreter a
      chance to e.g., print a prompt.  */
@@ -141,9 +142,12 @@ extern int current_interp_named_p (const char *name);
    interpreter should configure the output streams to send output only
    to the logfile.  If false, the interpreter should configure the
    output streams to send output to both the current output stream
-   (i.e., the terminal) and the log file.  */
+   (i.e., the terminal) and the log file.  DEBUG_REDIRECT is same as
+   LOGGING_REDIRECT, but for the value of "set logging debugredirect"
+   instead.  */
 extern void current_interp_set_logging (ui_file_up logfile,
-                                       bool logging_redirect);
+                                       bool logging_redirect,
+                                       bool debug_redirect);

 /* Returns the top-level interpreter.  */
 extern struct interp *top_level_interpreter (void);
diff --git a/gdb/mi/mi-interp.c b/gdb/mi/mi-interp.c
index 6a19bf0247..ad1a06cae0 100644
--- a/gdb/mi/mi-interp.c
+++ b/gdb/mi/mi-interp.c
@@ -1279,7 +1279,8 @@ mi_interp::interp_ui_out ()
    the consoles to use the supplied ui-file(s).  */

 void
-mi_interp::set_logging (ui_file_up logfile, bool logging_redirect)
+mi_interp::set_logging (ui_file_up logfile, bool logging_redirect,
+                       bool debug_redirect)
 {
   struct mi_interp *mi = this;

@@ -1289,22 +1290,30 @@ mi_interp::set_logging (ui_file_up logfile, bool logging_redirect)

       /* If something is being redirected, then grab logfile.  */
       ui_file *logfile_p = nullptr;
-      if (logging_redirect)
-       logfile_p = logfile.release ();
+      if (logging_redirect || debug_redirect)
+       {
+         logfile_p = logfile.get ();
+         mi->saved_raw_file_to_delete = logfile_p;
+       }

       /* If something is not being redirected, then a tee containing both the
         logfile and stdout.  */
       ui_file *tee = nullptr;
-      if (!logging_redirect)
-       tee = new tee_file (mi->raw_stdout, std::move (logfile));
+      if (!logging_redirect || !debug_redirect)
+       {
+         tee = new tee_file (mi->raw_stdout, std::move (logfile));
+         mi->saved_raw_file_to_delete = tee;
+       }

       mi->raw_stdout = logging_redirect ? logfile_p : tee;
+      mi->raw_stdlog = debug_redirect ? logfile_p : tee;
     }
   else
     {
-      delete mi->raw_stdout;
+      delete mi->saved_raw_file_to_delete;
       mi->raw_stdout = mi->saved_raw_stdout;
       mi->saved_raw_stdout = nullptr;
+      mi->saved_raw_file_to_delete = nullptr;
     }

   mi->out->set_raw (mi->raw_stdout);
diff --git a/gdb/testsuite/ChangeLog b/gdb/testsuite/ChangeLog
index 6d35156499..50888236af 100644
--- a/gdb/testsuite/ChangeLog
+++ b/gdb/testsuite/ChangeLog
@@ -1,3 +1,7 @@
+2019-05-17  Alan Hayward  <alan.hayward@arm.com>
+
+       * gdb.base/ui-redirect.exp: Add debug redirect tests.
+
 2019-05-17  Alan Hayward  <alan.hayward@arm.com>

        * gdb.base/ui-redirect.exp: Test redirection.
diff --git a/gdb/testsuite/gdb.base/ui-redirect.exp b/gdb/testsuite/gdb.base/ui-redirect.exp
index f1d00b939d..e62d5e8341 100644
--- a/gdb/testsuite/gdb.base/ui-redirect.exp
+++ b/gdb/testsuite/gdb.base/ui-redirect.exp
@@ -34,9 +34,17 @@ gdb_test_multiple $test $test {
 }
 gdb_test_no_output "end"

+if ![runto_main] {
+    fail "can't run to main"
+    return -1
+}
+gdb_breakpoint "foo"
+gdb_breakpoint "bar"
+
 with_test_prefix "logging" {
     gdb_test_no_output "set logging file /dev/null"
-    gdb_test "set logging on" "Copying output to /dev/null\\."
+    gdb_test "set logging on" \
+    "Copying output to /dev/null.*Copying debug output to /dev/null\\."
     gdb_test "save breakpoints /dev/null" "Saved to file '/dev/null'\\."
     gdb_test "set logging off" "Done logging to /dev/null\\."
     gdb_test "help" "List of classes of commands:.*"
@@ -44,7 +52,8 @@ with_test_prefix "logging" {

 with_test_prefix "redirect" {
     gdb_test "set logging redirect on"
-    gdb_test "set logging on" "Redirecting output to /dev/null\\."
+    gdb_test "set logging on" \
+    "Redirecting output to /dev/null.*Copying debug output to /dev/null\\."
     gdb_test_no_output "save breakpoints /dev/null"
     gdb_test "set logging off" "Done logging to /dev/null\\."
     gdb_test "help" "List of classes of commands:.*"
@@ -52,10 +61,33 @@ with_test_prefix "redirect" {

 with_test_prefix "redirect while already logging" {
     gdb_test_no_output "set logging redirect off"
-    gdb_test "set logging on" "Copying output to /dev/null\\."
+    gdb_test "set logging on" \
+    "Copying output to /dev/null.*Copying debug output to /dev/null\\."
     gdb_test "set logging redirect on" \
     ".*warning: Currently logging .*Turn the logging off and on to make the new setting effective.*"
     gdb_test "save breakpoints /dev/null" "Saved to file '/dev/null'\\."
     gdb_test "set logging off" "Done logging to /dev/null\\."
     gdb_test "help" "List of classes of commands:.*"
+    gdb_test_no_output "set logging redirect off"
+}
+
+with_test_prefix "debugging" {
+    gdb_test "set debug infrun 1"
+    gdb_test "set logging on" \
+    "Copying output to /dev/null.*Copying debug output to /dev/null\\."
+    gdb_test "continue" "Continuing.*infrun:.*infrun:.*Breakpoint \[0-9\]+, foo.*"
+    gdb_test "set debug infrun 0"
+    gdb_test "set logging off" "Done logging to /dev/null\\."
+    gdb_test "help" "List of classes of commands:.*"
+}
+
+with_test_prefix "redirect debugging" {
+    gdb_test_no_output "set logging debugredirect on"
+    gdb_test "set debug infrun 1"
+    gdb_test "set logging on" \
+    "Copying output to /dev/null.*Redirecting debug output to /dev/null\\."
+    gdb_test "continue" "Continuing.*((?!infrun).).*Breakpoint \[0-9\]+, bar.*"
+    gdb_test "set debug infrun 0"
+    gdb_test "set logging off" "Done logging to /dev/null\\."
+    gdb_test "help" "List of classes of commands:.*"
 }

diff --git a/gdb/mi/mi-common.h b/gdb/mi/mi-common.h
index 4fb2d75c0e..07b37b7144 100644
--- a/gdb/mi/mi-common.h
+++ b/gdb/mi/mi-common.h
@@ -66,7 +66,8 @@ public:
   void suspend () override;
   gdb_exception exec (const char *command_str) override;
   ui_out *interp_ui_out () override;
-  void set_logging (ui_file_up logfile, bool logging_redirect) override;
+  void set_logging (ui_file_up logfile, bool logging_redirect,
+                   bool debug_redirect) override;
   void pre_command_loop () override;

   /* MI's output channels */
@@ -79,9 +80,16 @@ public:
   /* Raw console output.  */
   struct ui_file *raw_stdout;

-  /* Save the original value of raw_stdout here when logging, so we
-     can restore correctly when done.  */
+  /* Raw logfile output.  */
+  struct ui_file *raw_stdlog;
+
+  /* Save the original value of raw_stdout and raw_stdlog here when logging, and
+     the file which we need to delete, so we can restore correctly when
+     done.  */
   struct ui_file *saved_raw_stdout;
+  struct ui_file *saved_raw_stdlog;
+  struct ui_file *saved_raw_file_to_delete;
+

   /* MI's builder.  */
   struct ui_out *mi_uiout;